;------------------------------------------------------------------------------
;	Fast Assembly MPEG Encoder
;------------------------------------------------------------------------------
; bitio.nas:
;	Performs output of data
;
;------------------------------------------------------------------------------
;
;	Copyright (C) 1999-2000 Vivien Chappelier
;
; *  This program is free software; you can redistribute it and/or modify
; *  it under the terms of the GNU General Public Licensse as published by
; *  the Free Software Foundation; either version 2 of the License, or
; *  (at your option) any later version.
; *
; *  This program is distributed in the hope that it will be useful,
; *  but WITHOUT ANY WARRANTY; without even the implied warranty of
; *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; *  GNU General Public License for more details.
; *
; *  You should have received a copy of the GNU General Public License
; *  along with this program; if not, write to the Free Software
; *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;
;------------------------------------------------------------------------------

; Macro definitions for portability

%macro cglobal 1
       global _%1
%define %1 _%1
%endmacro

%macro cextern 1
          extern _%1
%define %1 _%1
%endmacro

%ifdef _WIN32_
%define public cglobal
%endif

%ifdef _LINUX_
%define public global
%endif

; /////////////////////////////////////////////////////////////////////////////
; //                         	     DATA                                    //
; /////////////////////////////////////////////////////////////////////////////

section .data
alignb 4

       ; Why does this @=&"'& processor mask the five lowest bits of cl during shifts, shl ???, 32 does nothing :( 
        bitmasks   dd 00000000h
                   dd 00000001h
                   dd 00000003h
                   dd 00000007h
                   dd 0000000Fh
                   dd 0000001Fh
                   dd 0000003Fh
                   dd 0000007Fh
                   dd 000000FFh
                   dd 000001FFh
                   dd 000003FFh
                   dd 000007FFh
                   dd 00000FFFh
                   dd 00001FFFh
                   dd 00003FFFh
                   dd 00007FFFh
                   dd 0000FFFFh
                   dd 0001FFFFh
                   dd 0003FFFFh
                   dd 0007FFFFh
                   dd 000FFFFFh
                   dd 001FFFFFh
                   dd 003FFFFFh
                   dd 007FFFFFh
                   dd 00FFFFFFh
                   dd 01FFFFFFh
                   dd 03FFFFFFh
                   dd 07FFFFFFh
                   dd 0FFFFFFFh
                   dd 1FFFFFFFh
                   dd 3FFFFFFFh
                   dd 7FFFFFFFh
                   dd 0FFFFFFFFh
                   
; /////////////////////////////////////////////////////////////////////////////
; //                         	     CODE                                    //
; /////////////////////////////////////////////////////////////////////////////

section .text

        ; Bit input/output
        public BITWRITE
        public BITRESET

%include "io/bitwrite.inc"

; *****************************************************************************
; * BITWRITE function			                                      *
; *****************************************************************************
; * C prototype:							      *
; * 	void BITWRITE(bitbuffer_t * buffer, 				      *
; *			unsigned int data, unsigned int nbits);		      *
; *                                                                           *
; * Function : Writes nbits bits from data to output buffer		      *
; * 									      *
; * Assumptions: nbits <= 32						      *
; *									      *
; * Side Effects : None							      *
; *****************************************************************************

align 4
BITWRITE:
        push ebp
        mov ebp, esp
        push ebx

        mov   eax, [ebp + 12]              ; EAX = data
        mov   ecx, [ebp + 16]		   ; ECX = nbits
	mov   ebp, [ebp + 8]		   ; EBP = buffer
        BitWrite
        
        pop ebx
        pop ebp
        ret

; *****************************************************************************
; * BITRESET function			                                      *
; *****************************************************************************
; * C prototype:							      *
; * 	void BITRESET(bitbuffer_t * buffer)				      *
; *                                                                           *
; * Function : Clear the buffer						      *
; * 									      *
; * Side Effects : None							      *
; *****************************************************************************

align 4
BITRESET:
	push ebp
	mov ebp, esp
        push edi
	mov ebp, [ebp + 8]	; ebp = buffer
        mov edi, [ebp + 4]	; edi = buffer.data
        mov ecx, [ebp]		; ecx = buffer.nbits
        shr ecx, 5
        inc ecx
        mov eax, 0
align 4
bitresetl0:        
        mov [edi], eax
        add edi, 4
        dec ecx
        jnz bitresetl0
        mov [ebp], eax
	mov eax, ecx

        pop edi
	pop ebp
        ret
END
