;------------------------------------------------------------------------------
;	Fast Assembly MPEG Encoder
;------------------------------------------------------------------------------
; bitwrite.inc:
;	Macros for bitwise writing
;
;------------------------------------------------------------------------------
;
;	Copyright (C) 1999-2000 Vivien Chappelier
;
; *  This program is free software; you can redistribute it and/or modify
; *  it under the terms of the GNU General Public Licensse as published by
; *  the Free Software Foundation; either version 2 of the License, or
; *  (at your option) any later version.
; *
; *  This program is distributed in the hope that it will be useful,
; *  but WITHOUT ANY WARRANTY; without even the implied warranty of
; *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; *  GNU General Public License for more details.
; *
; *  You should have received a copy of the GNU General Public License
; *  along with this program; if not, write to the Free Software
; *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;
;------------------------------------------------------------------------------

; *****************************************************************************
; * BitWrite macro			                                      *
; *****************************************************************************
; *                                                                           *
; * Function : Writes ECX bits from EAX to output buffer pointed by EBP	      *
; *									      *
; * output buffer structure: dword bitnumber				      *
; * 			     dword buffer_pointer			      *
; *									      *
; * Side Effects : EAX, EBX, ECX, EDX registers modified		      *
; * 		   output_buffer modified				      *
; *****************************************************************************

%macro BitWrite 0

        mov   ebx, [ebp]
        add   [ebp], ecx
        mov   edx, 1
        dec   ecx
        shl   edx, cl
        shl   edx, 1
        dec   edx
        and   eax, edx		; mask EAX so that bits > nbits are cleared
        ror   eax, cl
        mov   edx, 0FFFFFFFFh
        mov   ecx, ebx
        shr   ebx, 5
        shr   edx, cl
        inc   cl
        ror   eax, cl        
        mov   ecx, eax        
        and   ecx, edx
        not   edx
        bswap ecx
        and   eax, edx
	mov   edx, [ebp + 4]
        or    dword [edx + 4*ebx], ecx
        bswap eax
        or    dword [edx + 4 + 4*ebx], eax

%endm

; *****************************************************************************
; * BitWriteNoMask macro		                                      *
; *****************************************************************************
; *                                                                           *
; * Function : Writes ECX bits from EAX to output buffer but without masking  *
; *	       EAX before writting. This is less secure than BitWrite but     *
; *	       it is faster.						      *
; *									      *
; * output buffer structure: dword bitnumber				      *
; * 			     dword buffer_pointer			      *
; *									      *
; * Side Effects : EAX, EBX, ECX, EDX registers modified		      *
; * 		   output_buffer modified				      *
; *****************************************************************************

%macro BitWriteNoMask 0

        mov   ebx, [ebp]
        add   [ebp], ecx
        mov   edx, 0FFFFFFFFh
        ror   eax, cl
        mov   ecx, ebx
        shr   ebx, 5
        shr   edx, cl
        ror   eax, cl        
        mov   ecx, eax
        and   eax, edx
        not   edx
        bswap eax
        and   ecx, edx
	mov   edx, [ebp + 4]
        or    dword [edx + 4*ebx], eax
        bswap ecx
        or    dword [edx + 4 + 4*ebx], ecx

%endm