/* 
 *    file.c
 *
 *	Copyright (C) Vivien Chappelier - 2000
 *
 *  This file is part of fame, a free MPEG encoder.
 *	
 *  fame is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  fame is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define MODULE_NAME "file"

char module_id[] = MODULE_NAME;

typedef struct _FileModule_ {
  int out;
  int limit;
  int total;
  int counter;
  char * filename;
} FileModule;

int is_sequence_header(unsigned char * data);

FileModule * module_init(int argc, char * argv[])
{
  FileModule * context;
  register int i;

  context = (FileModule *) malloc(sizeof(FileModule));
  context->out = -1;
  context->limit = 0;
  context->total = 0;
  context->counter = 0;
  context->filename = 0;

  for(i = 0; i < argc; i++)
  {
    if(!strncmp(argv[i], "module=", strlen("module=")))
    {
      if(strcmp(argv[i]+strlen("module="), MODULE_NAME)) return(0);
    }
    else
    if(!strcmp(argv[i], "help"))
    {
      fprintf(stderr, "   %s module options:\n", MODULE_NAME);
      fprintf(stderr, "      file=FILE           : set the output file\n");
      return(0);
    }
    else
    if(!strncmp(argv[i], "stdout", strlen("stdout")))
    {
      context->out = 1;
    }
    else
    if(!strncmp(argv[i], "file=", strlen("file=")))
    {
      char name[256];

      context->filename = (char *) malloc(256 * sizeof(char));
      strncpy(context->filename, argv[i]+strlen("file="), 256);
      sprintf(name, context->filename, context->counter);
      context->out = creat(name, 0644);
    }
    else
    if(!strncmp(argv[i], "limit=", strlen("limit=")))
    {
      context->limit = atoi(argv[i]+strlen("limit="));
    }
  }

  if(context->out == -1) return(0);

  return(context);
}

void module_close(FileModule * context)
{
  if(context)
  {
      if(context->out != -1) close(context->out);
      if(context->filename) free(context->filename);
      free(context);
  }
}

void output_write(FileModule * context, void * data, unsigned long size)
{
  if(context)
  {
    if(context->limit && context->filename != NULL && context->total > context->limit)
    {
      // Align on sequence start code
      if(is_sequence_header(data))
      {
	char name[256];

	close(context->out);
	context->counter++;
	sprintf(name, context->filename, context->counter);
	context->out = creat(name, 0644);
	context->total = 0;
      }
    }

    if(context->out != -1)
    {
      write(context->out, data, size);
      context->total += size;
    }
  }
}

int is_sequence_header(unsigned char * data)
{
  return(data[0] == 0x00 &&
	 data[1] == 0x00 &&
	 data[2] == 0x01 &&
	 data[3] == 0xb3);
}
