/* 
 *    network.cpp
 *
 *	Copyright (C) Vivien Chappelier - 2000
 *
 *  This file is part of fame, a free MPEG encoder.
 *	
 *  fame is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  fame is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sockstream.h>
#include <sockinet.h>

#include "sockmcast.h"

#define MODULE_NAME "network"

char module_id[] = MODULE_NAME;

struct NetworkModule {
  ostream * out;
};

bool is_address_multicast(sockinetaddr sa);

extern "C" NetworkModule * module_init(int argc, char * argv[])
{
  NetworkModule * context;
  register int i;
  char * address;
  int port;

  context = new NetworkModule;
  context->out = 0;
  address = 0;
  port = 0;

  for(i = 0; i < argc; i++)
  {
    if(!strncmp(argv[i], "module=", strlen("module=")))
    {
      if(strcmp(argv[i]+strlen("module="), MODULE_NAME)) return(0);
    }
    else
    if(!strcmp(argv[i], "help"))
    {
      fprintf(stderr, "   %s module options:\n", MODULE_NAME);
      fprintf(stderr, "      address=ADDRESS      : set the address of the receiver(s)\n");
      fprintf(stderr, "      port=PORT            : set the port\n");
      return(0);
    }
    else
    if(!strncmp(argv[i], "address=", strlen("address=")))
    {
      address = argv[i]+strlen("address=");
    }
    else
    if(!strncmp(argv[i], "port=", strlen("port=")))
    {
      port = atoi(argv[i]+strlen("port="));
    }
  }

  if(!address) return(false);

  sockinetaddr sa(address, port);

  if(is_address_multicast(sa))
  {
    context->out = new osockmcast;

    ((sockmcastbuf *) context->out->rdbuf())->reuseaddr(true);
    ((sockmcastbuf *) context->out->rdbuf())->linebuffered(0);
    ((sockmcastbuf *) context->out->rdbuf())->bind(sa);
    ((sockmcastbuf *) context->out->rdbuf())->connect(sa);
    
    /* Time to live is set to 1 */
    //	((sockmcastbuf *) context->out)->set_ttl(1);
  }
  else
  {
    context->out = new osockinet(sockbuf::sock_dgram);

    ((sockinetbuf *) context->out->rdbuf())->reuseaddr(true);
    ((sockinetbuf *) context->out->rdbuf())->linebuffered(0);
    ((sockinetbuf *) context->out->rdbuf())->bind();

    ((sockinetbuf *) context->out->rdbuf())->connect(sa);
  }

  if(!context->out || !context->out->good()) return(0);

  return(context);
}

extern "C" void module_close(NetworkModule * context)
{
  if(context && context->out) delete context->out;
  if(context) delete context;
}

extern "C" void output_write(NetworkModule * context, void * data, unsigned long size)
{
  if(context && context->out)
  {
    context->out->write(data, size);
    context->out->flush();

    /* Stop if disconnected */
    if(!context->out->good()) raise(SIGINT);
  }
}

bool is_address_multicast(sockinetaddr sa)
{
  long addr = ((struct sockaddr_in *) sa.addr())->sin_addr.s_addr;

  return(((addr & 255) >= 224 && (addr & 255) <= 239));
}
