
#include "xutil.h"

extern "C" {

/* Additional colormap support my M. Hoch */

int setup_xwinbuffer(struct xwinbuffer *xwin, char *prgname, int width, int height, int depth) {
  int screen;
  XWindowAttributes win_attributes;
  XVisualInfo desired_visual_template;
  XVisualInfo *visual_info_list; 
  int no_visual_matched;
  XGCValues gc_values;
  int localdepth;
  Colormap   wtcolormap ;
  XColor   * palette;
  int        pixelNum, numCols;

  /* Store parameters */
  xwin->width=width;
  xwin->height=height;
  xwin->depth=depth; 
  
  /* Open window */
  xwin->display_ = XOpenDisplay(NULL);
  xwin->wtwin_ = XCreateSimpleWindow(xwin->display_, DefaultRootWindow(xwin->display_),
			       0, 0, xwin->width, xwin->height, xwin->depth,
			       BlackPixel(xwin->display_, DefaultScreen(xwin->display_)),
			       BlackPixel(xwin->display_, DefaultScreen(xwin->display_)));  
  XMapWindow(xwin->display_, xwin->wtwin_);
  XSelectInput(xwin->display_, xwin->wtwin_, KeyPressMask | KeyReleaseMask | ButtonPressMask);
  gc_values.graphics_exposures = False;
  xwin->gc_ = XCreateGC(xwin->display_, xwin->wtwin_, GCGraphicsExposures, &gc_values);  

  /* Create shared memory segment for direct framebuffer access */
  screen = DefaultScreen(xwin->display_);
  XGetWindowAttributes(xwin->display_, xwin->wtwin_, &win_attributes);
  localdepth=win_attributes.depth;


  // check if desired depth and X-Server bpp match
  if ( depth == 1 && localdepth != 8 ) { // BW-Images 
   fprintf (stderr, "%%setup_xwinbuffer(): depth=1 requires X-Server in 8 bpp mode!\n"); 
   return(-1); 
  } 			   
  else if ( depth == 2 && localdepth != 16 ) { // RGB565-Images 
   fprintf (stderr, "%%setup_xwinbuffer(): depth=2 requires X-Server in 16 bpp mode!\n"); 
   return(-1); 
  } 			   
  else if ( depth > 2 ) {
   fprintf (stderr, "%%setup_xwinbuffer(): not implemented depth/X-Server mode combination (%d, %d)!\n", depth, localdepth); 
   return(-1); 
  } 			   

  xwin->shminfo = (XShmSegmentInfo*) malloc (sizeof(XShmSegmentInfo));
  xwin->shmimage = XShmCreateImage(xwin->display_, DefaultVisual(xwin->display_, screen),
			     localdepth, ZPixmap, NULL, xwin->shminfo,
			     xwin->width,xwin->height);

  if ((xwin->shminfo->shmid=shmget(getpid(),
			  xwin->shmimage->bytes_per_line * xwin->shmimage->height,
			  IPC_CREAT | IPC_EXCL | 0777))==-1) {
   perror("shmget()");
   return(-1);
  } 			  
  xwin->shminfo->shmaddr = (char *)shmat(xwin->shminfo->shmid, (void *) 0, 0);
  xwin->shmimage->data = xwin->shminfo->shmaddr;
  XShmAttach(xwin->display_, xwin->shminfo);

  // Set color mapping
  if ( depth == 1 ) // BW-Images 
    desired_visual_template.class  = PseudoColor;
  else
    desired_visual_template.class = TrueColor; 

  desired_visual_template.screen = screen;
  visual_info_list = XGetVisualInfo(xwin->display_,  
				    VisualScreenMask | VisualClassMask, 
				    &desired_visual_template, 
				    &no_visual_matched); 
  if (no_visual_matched == 0) {
   fprintf(stderr, "No visual matched\n");
   return(-1);
  }


  // make color map 
  if ( depth == 1 ) { // color map needed when using BW-Images 
    wtcolormap = XCreateColormap(xwin->display_, 
                                 DefaultRootWindow(xwin->display_),
                                 visual_info_list[0].visual, AllocAll);
    numCols = 1<<localdepth;
    palette = (XColor *)malloc(numCols * sizeof(XColor));
    for (pixelNum=0; pixelNum < numCols; pixelNum++) {
        palette[pixelNum].pixel  = pixelNum;
        palette[pixelNum].red    = pixelNum << (16-localdepth);
        palette[pixelNum].green  = pixelNum << (16-localdepth);
        palette[pixelNum].blue   = pixelNum << (16-localdepth);
        palette[pixelNum].flags  = DoRed|DoGreen|DoBlue;
      }

    XStoreName(xwin->display_, xwin->wtwin_, prgname);
    XStoreColors(xwin->display_, wtcolormap, palette, numCols);
    XSetWindowColormap(xwin->display_, xwin->wtwin_, wtcolormap);
  }
  
  return(0);  
}


int sync_xwinbuffer(struct xwinbuffer *xwin)
{
 /* Draw screen onto display */
 XShmPutImage(xwin->display_, xwin->wtwin_, xwin->gc_, xwin->shmimage,
 	        0, 0, 0, 0, xwin->width, xwin->height, False);
 XSync(xwin->display_, 0); 
 return(0);
}

int keypoll_xwinbuffer (struct xwinbuffer *xwin, KeySym *key, int *button, int *mousex, int *mousey)
{
  XEvent event; 
 
  /* Check keyboard/mouse input */
  if (XPending(xwin->display_)>0) {
   *button=0;
   *key=0;
   XNextEvent(xwin->display_, &event);
   switch (event.type) {
    case ButtonPress:
     *mousex=event.xbutton.x;
     *mousey=event.xbutton.y;
     *button=1;
     break;
    case KeyPress:
     *key=XLookupKeysym((void *)&event,0);
     break;
    case MappingNotify:
     XRefreshKeyboardMapping((void  *)&event);
     break;
   }
   return(1);
  } else { 
   return(0);
  } 
}  

int close_xwinbuffer (struct xwinbuffer *xwin)
{
 /* Mark Shm as destroyed */
 shmctl(xwin->shminfo->shmid,IPC_RMID,NULL);

 /* Detatch shared memory segment */
 shmdt ((char *)xwin->shminfo->shmaddr);
 
 return(0);
}

/* Draw horizontal line into color buffer */
void hline (struct xwinbuffer *xwin, int x1, int x2, int y, int color)
{
 int i;
 unsigned char *curpos8,color8;
 unsigned short *curpos16,color16;
 int dummy;
 
 if (xwin==NULL) return; 
 if (x2<x1) {
  dummy=x1;
  x1=x2;
  x2=dummy;
 }
 if (x1<0) {
  x1=0;
 } else if (x1>(xwin->width-1)) {
  x1=xwin->width-1;
 } 
 if (x2<0) {
  x2=0;
 } else if (x2>(xwin->width-1)) {
  x2=xwin->width-1;
 } 
 if (y<0) {
  y=0;
 } else if (y>(xwin->height-1)) {
  y=xwin->height-1;
 } 
 if (xwin->depth==1) {
  color8=(unsigned char)color;
  curpos8 = (unsigned char *)xwin->shmimage->data;
  curpos8 += (x1+y*xwin->width);
  for (i=0; i<(x2-x1+1); i++) {
   *curpos8=color8;
   curpos8++;
  }
 } else if (xwin->depth==2) {
  color16=(unsigned short)color;
  curpos16 = (unsigned short *)xwin->shmimage->data;
  curpos16 += (x1+y*xwin->width);
  for (i=0; i<=(x2-x1); i++) {
   *curpos16=color16;
   curpos16++;
  }
 }
}

/* Draw vertical line into color buffer */
void vline (struct xwinbuffer *xwin, int x, int y1, int y2, int color)
{
 int i;
 unsigned char *curpos8,color8;
 unsigned short *curpos16,color16;
 int dummy;

 if (xwin==NULL) return; 
 if (y2<y1) {
  dummy=y1;
  y1=y2;
  y2=dummy;
 }
 if (x<0) {
  x=0;
 } else if (x>(xwin->width-1)) {
  x=xwin->width-1;
 } 
 if (y1<0) {
  y1=0;
 } else if (y1>(xwin->height-1)) {
  y1=xwin->height-1;
 } 
 if (y2<0) {
  y2=0;
 } else if (y2>(xwin->height-1)) {
  y2=xwin->height-1;
 } 
 if (xwin->depth==1) {
  color8=(unsigned char)color;
  curpos8 = (unsigned char *)xwin->shmimage->data;
  curpos8 += (x+y1*xwin->width);
  for (i=y1; i<=y2; i++) {
   *curpos8=color8;
   curpos8 += xwin->width;
  }
 } else if (xwin->depth==2) {
  color16=(unsigned short)color;
  curpos16 = (unsigned short *)xwin->shmimage->data;
  curpos16 += (x+y1*xwin->width);
  for (i=y1; i<=y2; i++) {
   *curpos16=color16;
   curpos16 += xwin->width;
  }
 }
}


}
