;------------------------------------------------------------------------------
;	Fast Assembly MPEG Encoder
;------------------------------------------------------------------------------
; dct.nas:
;	Performs a 2D-DCT on the blocks using a dumb matrix product algorithm
;
; Notes: I'm almost convinced that other more sophisticated algorythm
;	 would take more time that this dumb one because of the time it takes
;	 to put data in the right registers.
;	 DCT needs to be done directly on input data to be fast.
;	 It is likely that Quantization could be done during the DCT 
;	 (what can't be done when using Fast DCT algorythm) if anyone want to
;	 try...well, try it.
;------------------------------------------------------------------------------
;
;	Copyright (C) 1999-2000 Vivien Chappelier
;
; *  This program is free software; you can redistribute it and/or modify
; *  it under the terms of the GNU General Public Licensse as published by
; *  the Free Software Foundation; either version 2 of the License, or
; *  (at your option) any later version.
; *
; *  This program is distributed in the hope that it will be useful,
; *  but WITHOUT ANY WARRANTY; without even the implied warranty of
; *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; *  GNU General Public License for more details.
; *
; *  You should have received a copy of the GNU General Public License
; *  along with this program; if not, write to the Free Software
; *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;
;------------------------------------------------------------------------------

; Macro definitions for portability
	
%macro cglobal 1 
       global _%1 
%define %1 _%1 
%endmacro

%macro cextern 1 
          extern _%1 
%define %1 _%1 
%endmacro

%ifdef _WIN32_
%define public cglobal
%endif

%ifdef _LINUX_
%define public global
%endif

; /////////////////////////////////////////////////////////////////////////////
; //                         	     DATA                                    //
; /////////////////////////////////////////////////////////////////////////////

section .data

        DCTSIZE equ 8


alignb 16
	; temporary 8x8 block
	dcttemp   times 8*8 dw 0
        
	; constants
        dctround    dd      00008000h
	            dd	    00008000h
        dctzero     dd             0
		    dd	           0
        dct128      dd      00800080h
		    dd      00800080h
	YUYV_maskY  dw 	    0FF00h, 0FF00h, 0FF00h, 0FF00h

	; dctmatrix(x,y) = 1/2*sin(A(x,y)/16*Pi);      
	;        where A is
	; [ 4  4  4  4  4  4  4  4 ]
	; [ 7  5  3  1 -1 -3 -5 -7 ]
	; [ 6  2 -2 -6 -6 -2  2  6 ]
	; [ 5 -1 -7 -3  3  7  1 -5 ]
	; [ 4 -4 -4  4  4 -4 -4  4 ]
	; [ 3 -7  1  5 -5 -1  7 -3 ]
	; [ 2 -6  6 -2 -2  6 -6  2 ]
	; [ 1 -3  5 -7  7 -5  3 -1 ]

alignb 16
        dctmatrix dw 05A82h,05A82h,05A82h,05A82h,05A82h,05A82h,05A82h,05A82h
                  dw 0E707h,0B8E3h,09592h,08276h,07D8Ah,06A6Eh,0471Dh,018F9h
                  dw 089BEh,0CF04h,030FCh,07642h,07642h,030FCh,0CF04h,089BEh
                  dw 0471Dh,07D8Ah,018F9h,09592h,06A6Eh,0E707h,08276h,0B8E3h
                  dw 05A82h,0A57Eh,0A57Eh,05A82h,05A82h,0A57Eh,0A57Eh,05A82h
                  dw 09592h,0E707h,07D8Ah,0B8E3h,0471Dh,08276h,018F9h,06A6Eh
                  dw 0CF04h,07642h,089BEh,030FCh,030FCh,089BEh,07642h,0CF04h
                  dw 07D8Ah,09592h,0471Dh,0E707h,018F9h,0B8E3h,06A6Eh,08276h

; /////////////////////////////////////////////////////////////////////////////
; //                         	     CODE                                    //
; /////////////////////////////////////////////////////////////////////////////

section .text
        
        ; The Dicrete Cosine Transform

	; for YUV9 planar format
        public DCTY_YUV9
        public DCTC_YUV9

	; for YUV12 planar format (fastest)
        public DCTY_YUV12
        public DCTC_YUV12

	; for YUYV format
        public DCTY_YUYV
        public DCTC_YUYV

	public DCT_TEST

; *****************************************************************************
; * OneLineDCT macro			                                      *
; *****************************************************************************
; * Function : Read a column from input and write to a column of output       *
; *            performing a one dimensional DCT                               *
; *									      *
; * Side Effects : MMX registers modified				      *
; *****************************************************************************

%macro OneLineDCT 1               ; (Aij)0<=i,j<=7 are the coeffs of the matrix
				  ; (Iij) the coeffs of the input block
				  ; 68 instructions
	
	; first line * first matrix row
        movq mm0, [esi + 8]           ; MM0 : [ I07 ][ I06 ][ I05 ][ I04 ]
        movq mm1, [esi]               ; MM1 : [ I03 ][ I02 ][ I01 ][ I00 ]
				      
        pmaddwd mm0, [dctmatrix+%1]   ; MM0 : [ I07*A07 + I06*A06 ] (msdw)
				      ;       [ I05*A05 + I04*A04 ] (lsdw)
        pmaddwd mm1, [dctmatrix+8+%1] ; MM1 : [ I03*A03 + I02*A02 ] (msdw)
				      ;	      [ I01*A01 + I00*A00 ] (lsdw)

	; second line * first matrix row
        movq mm2, [esi + 24]          ; MM2 : [ I17 ][ I16 ][ I15 ][ I14 ]
        movq mm3, [esi + 16]	      ; MM3 : [ I13 ][ I12 ][ I11 ][ I10 ]
        pmaddwd mm2, [dctmatrix+%1]   ; MM2 : [ I17*A07 + I16*A06 ] (msdw)
				      ;       [ I15*A05 + I14*A04 ] (lsdw)
        pmaddwd mm3, [dctmatrix+8+%1] ; MM3 : [ I13*A03 + I12*A02 ] (msdw)
				      ;       [ I11*A01 + I10*A00 ] (lsdw)

        paddd mm0, mm1                ; MM0 : [I07*A07+I06*A06+I03*A03+I02*A02]
				      ;	      [I05*A05+I04*A04+I01*A01+I00*A00]
        paddd mm2, mm3                ; MM2 : [I17*A07+I16*A06+I13*A03+I12*A02]
				      ;	      [I15*A05+I14*A04+I11*A01+I10*A00]

	; third line * first matrix row
        movq mm4, [esi + 40]          
        movq mm5, [esi + 32]
        pmaddwd mm4, [dctmatrix+%1]
        pmaddwd mm5, [dctmatrix+8+%1]

	; fourth line * first matrix row
        movq mm6, [esi + 56]          
        movq mm7, [esi + 48]
        pmaddwd mm6, [dctmatrix+%1]
        pmaddwd mm7, [dctmatrix+8+%1]

        paddd mm4, mm5                ; MM4 : [I27*A07+I26*A06+I23*A03+I22*A02]
				      ;	      [I25*A05+I24*A04+I21*A01+I20*A00]
        paddd mm6, mm7                ; MM6 : [I37*A07+I36*A06+I33*A03+I32*A02]
				      ;	      [I35*A05+I34*A04+I31*A01+I30*A00]
               
                                      ; I had to interlace the instruction so 
				      ;	as to optimize for Pentium processor
                                      ; as if you have MMX you generaly have 
				      ; at least a Pentium processor.
                                      ; The drawback is that the code loses 
				      ;	clarity :)
        movq mm1, mm0                 ; Now odd MMX register are free
        movq mm3, mm4                   
        punpckhdq mm0, mm2            ; MM0 : [I17*A07+I16*A06+I13*A03+I12*A02]
				      ;	      [I07*A07+I06*A06+I03*A03+I02*A02]
        punpckhdq mm4, mm6
        punpckldq mm3, mm6
        punpckldq mm1, mm2            ; MM1 : [I15*A05+I14*A04+I11*A01+I10*A00]
				      ;	      [I05*A05+I04*A04+I01*A01+I00*A00]
        paddd mm0, mm1                ; MM0 : [ Sum(i=0, 7, A0i*I1i) ] (msdw)
				      ;       [ Sum(i=0, 7, A0i*I0i) ] (lsdw)
        paddd mm4, mm3                ; MM4 : [ Sum(i=0, 7, A0i*I3i) ] (msdw)
				      ;       [ Sum(i=0, 7, A0i*I2i) ] (lsdw)
	paddd mm0, [dctround]         ; Round to nearest
        paddd mm4, [dctround]         ; Round to nearest
        psrad mm0, 16                 ; Remove Fixed point
        psrad mm4, 16
        
        packssdw mm0, mm4             ; MM0 : [ Sum(i=0, 7, A0i*I3i) ] (msw)
				      ;       [ Sum(i=0, 7, A0i*I2i) ]
				      ;       [ Sum(i=0, 7, A0i*I1i) ]
				      ;       [ Sum(i=0, 7, A0i*I0i) ] (lsw)

	; The result is stored in half a line of output
	; we just have to repeat the process
        movq [edi + %1], mm0      

	; fifth line * first matrix row
        movq mm0, [esi + 72]            
        movq mm1, [esi + 64]                 
        pmaddwd mm0, [dctmatrix+%1]         
        pmaddwd mm1, [dctmatrix+8+%1]
        paddd mm0, mm1                  

	; sixth line * first matrix row
        movq mm2, [esi + 88]           
        movq mm3, [esi + 80]
        pmaddwd mm2, [dctmatrix+%1]
        pmaddwd mm3, [dctmatrix+8+%1]
        paddd mm2, mm3                  

	; seventh line * first matrix row
        movq mm4, [esi + 104]          
        movq mm5, [esi + 96]
        pmaddwd mm4, [dctmatrix+%1]
        pmaddwd mm5, [dctmatrix+8+%1]
        paddd mm4, mm5                  

	; eightth line * first matrix row
        movq mm6, [esi + 120]           
        movq mm7, [esi + 112]
        pmaddwd mm6, [dctmatrix+%1]
        pmaddwd mm7, [dctmatrix+8+%1]
        paddd mm6, mm7                  
                 
        movq mm1, mm0                   
        punpckhdq mm0, mm2              
        punpckldq mm1, mm2              
        paddd mm0, mm1                  
        paddd mm0, [dctround]         ; Round to nearest
        psrad mm0, 16                 ; Remove Fixed point

        movq mm3, mm4                   
        punpckhdq mm4, mm6
        punpckldq mm3, mm6
        paddd mm4, mm3                  
        paddd mm4, [dctround]         ; Round to nearest        
        psrad mm4, 16
        
        packssdw mm0, mm4               

	; The result is stored in half a line of output
        movq [edi + 8 + %1], mm0  

%endm


; *****************************************************************************
; * OneLineDCTYExpand_Planar macro	                               	      *
; *****************************************************************************
; * Function : Read a column from input (bytes), unpack it to words and write *
; *            to a column of output performing a one dimensional DCT.        *
; *	       See OneLineDCT for comments.				      *
; *									      *
; * Side Effects : MMX registers modified				      *
; *****************************************************************************

%macro OneLineDCTYExpand_Planar 2

	; first line * first matrix row
        push esi
        movq mm0, [esi]                 
        movq mm1, [esi]
        punpckhbw mm0, [dctzero]
        punpcklbw mm1, [dctzero]	
        pmaddwd mm0, [dctmatrix+%1]   	
					
        pmaddwd mm1, [dctmatrix+8+%1] 	
					
        add esi, %2

	; second line * first matrix row
        movq mm2, [esi]                 
        punpckhbw mm2, [dctzero]        
        movq mm3, [esi]
        punpcklbw mm3, [dctzero]        
        pmaddwd mm2, [dctmatrix+%1]
        pmaddwd mm3, [dctmatrix+8+%1]
        add esi, %2

        paddd mm0, mm1                  
        paddd mm2, mm3                  

	; third line * first matrix row
        movq mm4, [esi]                 
        punpckhbw mm4, [dctzero]        
        movq mm5, [esi]
        punpcklbw mm5, [dctzero]        
        pmaddwd mm4, [dctmatrix+%1]
        pmaddwd mm5, [dctmatrix+8+%1]
        add esi, %2
        
	; fourth line * first matrix row
        movq mm6, [esi]                 
        punpckhbw mm6, [dctzero]        
        movq mm7, [esi]
        punpcklbw mm7, [dctzero]
        pmaddwd mm6, [dctmatrix+%1]
        pmaddwd mm7, [dctmatrix+8+%1]
        add esi, %2

        paddd mm4, mm5
        paddd mm6, mm7 

        movq mm1, mm0                   
        movq mm3, mm4                   
        punpckhdq mm0, mm2              
        punpckhdq mm4, mm6
        punpckldq mm3, mm6
        punpckldq mm1, mm2              
        paddd mm0, mm1                  
        paddd mm4, mm3                  
        paddd mm0, [dctround]
        paddd mm4, [dctround]
        psrad mm0, 16
        psrad mm4, 16
        
        packssdw mm0, mm4

        movq [edi + %1], mm0

        ; fifth line * first matrix row
        movq mm0, [esi]
        punpckhbw mm0, [dctzero]
        movq mm1, [esi]
        punpcklbw mm1, [dctzero]        
        pmaddwd mm0, [dctmatrix+%1]         
        pmaddwd mm1, [dctmatrix+8+%1]
        paddd mm0, mm1    
        add esi, %2              
        
	; sixth line * first matrix row
        movq mm2, [esi]
        punpckhbw mm2, [dctzero]        
        movq mm3, [esi]
        punpcklbw mm3, [dctzero]
        pmaddwd mm2, [dctmatrix+%1]
        pmaddwd mm3, [dctmatrix+8+%1]
        paddd mm2, mm3    
        add esi, %2              

	; seventh line * first matrix row
        movq mm4, [esi]
        punpckhbw mm4, [dctzero]        
        movq mm5, [esi]
        punpcklbw mm5, [dctzero]        
        pmaddwd mm4, [dctmatrix+%1]
        pmaddwd mm5, [dctmatrix+8+%1]
        paddd mm4, mm5    
        add esi, %2              
        
        ; eightth line * first matrix row
        movq mm6, [esi]
        punpckhbw mm6, [dctzero]        
        movq mm7, [esi]
        punpcklbw mm7, [dctzero]        
        pmaddwd mm6, [dctmatrix+%1]
        pmaddwd mm7, [dctmatrix+8+%1]
        paddd mm6, mm7    
                
        movq mm1, mm0                   
        punpckhdq mm0, mm2              
        punpckldq mm1, mm2              
        paddd mm0, mm1                  
        paddd mm0, [dctround]
        psrad mm0, 16

        movq mm3, mm4                   
        punpckhdq mm4, mm6
        punpckldq mm3, mm6
        paddd mm4, mm3                  
        paddd mm4, [dctround]
        psrad mm4, 16
        
        packssdw mm0, mm4               

        movq [edi + 8 + %1], mm0
        pop esi

%endm

; *****************************************************************************
; * OneLineDCTYExpand_YUYV macro	                               	      *
; *****************************************************************************
; * Function : Read a column from input (bytes), unpack it to words and write *
; *            to a column of output performing a one dimensional DCT.        *
; *	       See OneLineDCT for comments.				      *
; *									      *
; * Side Effects : MMX registers modified				      *
; *****************************************************************************

%macro OneLineDCTYExpand_YUYV 2

	; first line * first matrix row
        push esi
        movq mm0, [esi+8]                 
        movq mm1, [esi]
	pand mm0, [YUYV_maskY]
	pand mm1, [YUYV_maskY]
        psrlw mm0, 8
        psrlw mm1, 8
        pmaddwd mm0, [dctmatrix+%1]   	
        pmaddwd mm1, [dctmatrix+8+%1] 	
					
        add esi, %2

	; second line * first matrix row
        movq mm2, [esi+8]                 
        movq mm3, [esi]
	pand mm2, [YUYV_maskY]
	pand mm3, [YUYV_maskY]
        psrlw mm2, 8
        psrlw mm3, 8
        pmaddwd mm2, [dctmatrix+%1]
        pmaddwd mm3, [dctmatrix+8+%1]
        add esi, %2

        paddd mm0, mm1                  
        paddd mm2, mm3                  

	; third line * first matrix row
        movq mm4, [esi+8]                 
        movq mm5, [esi]
	pand mm4, [YUYV_maskY]
	pand mm5, [YUYV_maskY]
        psrlw mm4, 8
        psrlw mm5, 8
        pmaddwd mm4, [dctmatrix+%1]
        pmaddwd mm5, [dctmatrix+8+%1]
        add esi, %2
        
	; fourth line * first matrix row
        movq mm6, [esi+8]                 
        movq mm7, [esi]
	pand mm6, [YUYV_maskY]
	pand mm7, [YUYV_maskY]
        psrlw mm6, 8
        psrlw mm7, 8
        pmaddwd mm6, [dctmatrix+%1]
        pmaddwd mm7, [dctmatrix+8+%1]
        add esi, %2

        paddd mm4, mm5
        paddd mm6, mm7 

        movq mm1, mm0                   
        movq mm3, mm4                   
        punpckhdq mm0, mm2              
        punpckhdq mm4, mm6
        punpckldq mm3, mm6
        punpckldq mm1, mm2              
        paddd mm0, mm1                  
        paddd mm4, mm3                  
        paddd mm0, [dctround]
        paddd mm4, [dctround]
        psrad mm0, 16
        psrad mm4, 16
        
        packssdw mm0, mm4

        movq [edi + %1], mm0

        ; fifth line * first matrix row
        movq mm0, [esi+8]
        movq mm1, [esi]
	pand mm0, [YUYV_maskY]
	pand mm1, [YUYV_maskY]
        psrlw mm0, 8
        psrlw mm1, 8
        pmaddwd mm0, [dctmatrix+%1]         
        pmaddwd mm1, [dctmatrix+8+%1]
        paddd mm0, mm1    
        add esi, %2              
        
	; sixth line * first matrix row
        movq mm2, [esi+8]
        movq mm3, [esi]
	pand mm2, [YUYV_maskY]
	pand mm3, [YUYV_maskY]
        psrlw mm2, 8
        psrlw mm3, 8
        pmaddwd mm2, [dctmatrix+%1]
        pmaddwd mm3, [dctmatrix+8+%1]
        paddd mm2, mm3    
        add esi, %2              

	; seventh line * first matrix row
        movq mm4, [esi+8]
        movq mm5, [esi]
	pand mm4, [YUYV_maskY]
	pand mm5, [YUYV_maskY]
        psrlw mm4, 8
        psrlw mm5, 8
        pmaddwd mm4, [dctmatrix+%1]
        pmaddwd mm5, [dctmatrix+8+%1]
        paddd mm4, mm5    
        add esi, %2              
        
        ; eightth line * first matrix row
        movq mm6, [esi+8]
        movq mm7, [esi]
	pand mm6, [YUYV_maskY]
	pand mm7, [YUYV_maskY]
        psrlw mm6, 8
        psrlw mm7, 8
        pmaddwd mm6, [dctmatrix+%1]
        pmaddwd mm7, [dctmatrix+8+%1]
        paddd mm6, mm7    
                
        movq mm1, mm0                   
        punpckhdq mm0, mm2              
        punpckldq mm1, mm2              
        paddd mm0, mm1                  
        paddd mm0, [dctround]
        psrad mm0, 16

        movq mm3, mm4                   
        punpckhdq mm4, mm6
        punpckldq mm3, mm6
        paddd mm4, mm3                  
        paddd mm4, [dctround]
        psrad mm4, 16
        
        packssdw mm0, mm4               

        movq [edi + 8 + %1], mm0
        pop esi

%endm

; *****************************************************************************
; * OneLineDCTCExpand_12 macro		                                      *
; *****************************************************************************
; * Function : Read a column from input (bytes), unpack it to words and write *
; *            to a column of output performing a one dimensional DCT.        *
; *	       It does the same as OneLineDCTExpand except that it substracts *
; *            128 from the input (Cr and Cb components range from -128 to    *
; *            +127 whereas Y component ranges from 0 to 255)		      *
; * 	       This one is for YUV9 input ( where Cb and Cr are subsampled)   *
; *	       See OneLineDCT for comments.				      *
; *									      *
; * Side Effects : MMX registers modified				      *
; *****************************************************************************

%macro OneLineDCTCExpand_YUV12 2

        push esi
        ; first line * first matrix row
        movq mm0, [esi]                 
        movq mm1, [esi]
        punpckhbw mm0, [dctzero]
        punpcklbw mm1, [dctzero]
        psubw   mm0, [dct128]
        psubw   mm1, [dct128]
	pmaddwd mm0, [dctmatrix+%1]
        pmaddwd mm1, [dctmatrix+8+%1]
        add esi, %2

        ; second line * first matrix row
        movq mm2, [esi]
        movq mm3, [esi]
        punpckhbw mm2, [dctzero]
        punpcklbw mm3, [dctzero]
        psubw   mm2, [dct128]
        psubw   mm3, [dct128]        
        pmaddwd mm2, [dctmatrix+%1]
        pmaddwd mm3, [dctmatrix+8+%1]
        add esi, %2

        paddd mm0, mm1
        paddd mm2, mm3

        ; third line * first matrix row
        movq mm4, [esi]
        punpckhbw mm4, [dctzero]        
        movq mm5, [esi]
        punpcklbw mm5, [dctzero]
        psubw   mm4, [dct128]
        psubw   mm5, [dct128]
        pmaddwd mm4, [dctmatrix+%1]
        pmaddwd mm5, [dctmatrix+8+%1]
        add esi, %2
        
        ; fourth line * first matrix row
        movq mm6, [esi]
        punpckhbw mm6, [dctzero]        
        movq mm7, [esi]
        punpcklbw mm7, [dctzero]
        psubw   mm6, [dct128]
        psubw   mm7, [dct128]
        pmaddwd mm6, [dctmatrix+%1]
        pmaddwd mm7, [dctmatrix+8+%1]
        add esi, %2

        paddd mm4, mm5
        paddd mm6, mm7

        movq mm1, mm0
        movq mm3, mm4                   
        punpckhdq mm0, mm2
        punpckhdq mm4, mm6
        punpckldq mm3, mm6
        punpckldq mm1, mm2
        paddd mm0, mm1
        paddd mm4, mm3
        paddd mm0, [dctround]
        paddd mm4, [dctround]
        psrad mm0, 16
        psrad mm4, 16
        
        packssdw mm0, mm4

        movq [edi + %1], mm0

        ; fifth line * first matrix row
        movq mm0, [esi]
        punpckhbw mm0, [dctzero]
        movq mm1, [esi]
        punpcklbw mm1, [dctzero]
        psubw   mm0, [dct128]
        psubw   mm1, [dct128]
        pmaddwd mm0, [dctmatrix+%1]         
        pmaddwd mm1, [dctmatrix+8+%1]
        paddd mm0, mm1    
        add esi, %2              
        
        ; sixth line * first matrix row
        movq mm2, [esi]
        punpckhbw mm2, [dctzero]        
        movq mm3, [esi]
        punpcklbw mm3, [dctzero]
        psubw   mm2, [dct128]
        psubw   mm3, [dct128]
        pmaddwd mm2, [dctmatrix+%1]
        pmaddwd mm3, [dctmatrix+8+%1]
        paddd mm2, mm3    
        add esi, %2              

        ; seventh line * first matrix row
        movq mm4, [esi]
        punpckhbw mm4, [dctzero]        
        movq mm5, [esi]
        punpcklbw mm5, [dctzero]
        psubw   mm4, [dct128]
        psubw   mm5, [dct128]
        pmaddwd mm4, [dctmatrix+%1]
        pmaddwd mm5, [dctmatrix+8+%1]
        paddd mm4, mm5    
        add esi, %2              
        
        ; eightth line * first matrix row
        movq mm6, [esi]
        punpckhbw mm6, [dctzero]        
        movq mm7, [esi]
        punpcklbw mm7, [dctzero]
	psubw   mm6, [dct128]
        psubw   mm7, [dct128]
        pmaddwd mm6, [dctmatrix+%1]
        pmaddwd mm7, [dctmatrix+8+%1]
        paddd mm6, mm7    
                
        movq mm1, mm0                   
        punpckhdq mm0, mm2              
        punpckldq mm1, mm2              
        paddd mm0, mm1                  
        paddd mm0, [dctround]
        psrad mm0, 16

        movq mm3, mm4                   
        punpckhdq mm4, mm6
        punpckldq mm3, mm6
        paddd mm4, mm3                  
        paddd mm4, [dctround]
        psrad mm4, 16
        
        packssdw mm0, mm4               

        movq [edi + 8 + %1], mm0
        pop esi

%endm

; *****************************************************************************
; * OneLineDCTCExpand_YUV9 macro	                                      *
; *****************************************************************************
; * Function : Read a column from input (bytes), unpack it to words and write *
; *            to a column of output performing a one dimensional DCT.        *
; *	       It does the same as OneLineDCTExpand except that it substracts *
; *            128 from the input (Cr and Cb components range from -128 to    *
; *            +127 whereas Y component ranges from 0 to 255)		      *
; * 	       This one is for YUV9 input ( where Cb and Cr are subsampled)   *
; *	       See OneLineDCT for comments.				      *
; *									      *
; * Side Effects : MMX registers modified				      *
; *****************************************************************************

%macro OneLineDCTCExpand_YUV9 2

        push esi

	; first line * first matrix row
        movq mm0, [esi]
        punpcklbw mm0, [dctzero]              
        movq mm1, mm0
        punpckhwd mm0, mm1
        punpcklwd mm1, mm1
        psubw   mm0, [dct128]
        psubw   mm1, [dct128]        
        pmaddwd mm0, [dctmatrix+%1]
        pmaddwd mm1, [dctmatrix+8+%1]

	; second line * first matrix row
        movq mm2, [esi]
        punpcklbw mm2, [dctzero]
        movq mm3, mm2
        punpckhwd mm2, mm3
        punpcklwd mm3, mm3
        psubw   mm2, [dct128]
        psubw   mm3, [dct128]        
        pmaddwd mm2, [dctmatrix+%1]
        pmaddwd mm3, [dctmatrix+8+%1]
        add esi, %2

        paddd mm0, mm1
        paddd mm2, mm3
                      
	; third line * first matrix row
        movq mm4, [esi]
        punpcklbw mm4, [dctzero]        
        movq mm5, mm4        
        punpckhwd mm4, mm5
        punpcklwd mm5, mm5
        psubw   mm4, [dct128]
        psubw   mm5, [dct128]        
        pmaddwd mm4, [dctmatrix+%1]
        pmaddwd mm5, [dctmatrix+8+%1]
        
        ; fourth line * first matrix row
        movq mm6, [esi]
        punpcklbw mm6, [dctzero]
        movq mm7, mm6        
        punpckhwd mm6, mm7
        punpcklwd mm7, mm7
        psubw   mm6, [dct128]
        psubw   mm7, [dct128]        
        pmaddwd mm6, [dctmatrix+%1]
        pmaddwd mm7, [dctmatrix+8+%1]
        add esi, %2

        paddd mm4, mm5
        paddd mm6, mm7
                
        movq mm1, mm0
        movq mm3, mm4                   
        punpckhdq mm0, mm2
        punpckhdq mm4, mm6
        punpckldq mm3, mm6
        punpckldq mm1, mm2
        paddd mm0, mm1
        paddd mm4, mm3
        paddd mm0, [dctround]
        paddd mm4, [dctround]
        psrad mm0, 16
        psrad mm4, 16    
    
        packssdw mm0, mm4

        movq [edi + %1], mm0

        ; fifth line * first matrix row
        movq mm0, [esi]
        punpcklbw mm0, [dctzero]        
        movq mm1, mm0       
        punpckhwd mm0, mm1
        punpcklwd mm1, mm1
        psubw   mm0, [dct128]
        psubw   mm1, [dct128]        
        pmaddwd mm0, [dctmatrix+%1]         
        pmaddwd mm1, [dctmatrix+8+%1]
        paddd mm0, mm1    
        
        ; sixth line * first matrix row
        movq mm2, [esi]
        punpcklbw mm2, [dctzero]
        movq mm3, mm2
        punpckhwd mm2, mm3
        punpcklwd mm3, mm3
        psubw   mm2, [dct128]
        psubw   mm3, [dct128]        
        pmaddwd mm2, [dctmatrix+%1]
        pmaddwd mm3, [dctmatrix+8+%1]
        paddd mm2, mm3    
        add esi, %2              

	; seventh line * first matrix row
        movq mm4, [esi]
        punpcklbw mm4, [dctzero]
        movq mm5, mm4
        punpckhwd mm4, mm5
        punpcklwd mm5, mm5
        psubw   mm4, [dct128]
        psubw   mm5, [dct128]        
        pmaddwd mm4, [dctmatrix+%1]
        pmaddwd mm5, [dctmatrix+8+%1]
        paddd mm4, mm5    
        
        ; eightth line * first matrix row
        movq mm6, [esi]
        punpcklbw mm6, [dctzero]
        movq mm7, mm6
        punpckhwd mm6, mm7
        punpcklwd mm7, mm7
        psubw   mm6, [dct128]
        psubw   mm7, [dct128]        
        pmaddwd mm6, [dctmatrix+%1]
        pmaddwd mm7, [dctmatrix+8+%1]
        paddd mm6, mm7    
                
        movq mm1, mm0                   
        punpckhdq mm0, mm2              
        punpckldq mm1, mm2              
        paddd mm0, mm1                  
        paddd mm0, [dctround]
        psrad mm0, 16

        movq mm3, mm4                   
        punpckhdq mm4, mm6
        punpckldq mm3, mm6
        paddd mm4, mm3                  
        paddd mm4, [dctround]
        psrad mm4, 16
        
        packssdw mm0, mm4               

        movq [edi + 8 + %1], mm0

        pop esi

%endm

; *****************************************************************************
; * OneLineDCTCExpand_YUYV macro	                                      *
; *****************************************************************************
; * Function : Read a column from input (bytes), unpack it to words and write *
; *            to a column of output performing a one dimensional DCT.        *
; *	       It does the same as OneLineDCTExpand except that it substracts *
; *            128 from the input (Cr and Cb components range from -128 to    *
; *            +127 whereas Y component ranges from 0 to 255)		      *
; * 	       This one is for YUV9 input ( where Cb and Cr are subsampled)   *
; *	       See OneLineDCT for comments.				      *
; *									      *
; * Side Effects : MMX registers modified				      *
; *****************************************************************************

%macro OneLineDCTCExpand_YUYV 2

        push esi
        ; first line * first matrix row
        movq mm0, [esi]                 
        movq mm1, [esi]
        punpckhbw mm0, [dctzero]
        punpcklbw mm1, [dctzero]
        psubw   mm0, [dct128]
        psubw   mm1, [dct128]
	pmaddwd mm0, [dctmatrix+%1]
        pmaddwd mm1, [dctmatrix+8+%1]
        add esi, %2

        ; second line * first matrix row
        movq mm2, [esi]
        movq mm3, [esi]
        punpckhbw mm2, [dctzero]
        punpcklbw mm3, [dctzero]
        psubw   mm2, [dct128]
        psubw   mm3, [dct128]        
        pmaddwd mm2, [dctmatrix+%1]
        pmaddwd mm3, [dctmatrix+8+%1]
        add esi, %2

        paddd mm0, mm1
        paddd mm2, mm3

        ; third line * first matrix row
        movq mm4, [esi]
        punpckhbw mm4, [dctzero]        
        movq mm5, [esi]
        punpcklbw mm5, [dctzero]
        psubw   mm4, [dct128]
        psubw   mm5, [dct128]
        pmaddwd mm4, [dctmatrix+%1]
        pmaddwd mm5, [dctmatrix+8+%1]
        add esi, %2
        
        ; fourth line * first matrix row
        movq mm6, [esi]
        punpckhbw mm6, [dctzero]        
        movq mm7, [esi]
        punpcklbw mm7, [dctzero]
        psubw   mm6, [dct128]
        psubw   mm7, [dct128]
        pmaddwd mm6, [dctmatrix+%1]
        pmaddwd mm7, [dctmatrix+8+%1]
        add esi, %2

        paddd mm4, mm5
        paddd mm6, mm7

        movq mm1, mm0
        movq mm3, mm4                   
        punpckhdq mm0, mm2
        punpckhdq mm4, mm6
        punpckldq mm3, mm6
        punpckldq mm1, mm2
        paddd mm0, mm1
        paddd mm4, mm3
        paddd mm0, [dctround]
        paddd mm4, [dctround]
        psrad mm0, 16
        psrad mm4, 16
        
        packssdw mm0, mm4

        movq [edi + %1], mm0

        ; fifth line * first matrix row
        movq mm0, [esi]
        punpckhbw mm0, [dctzero]
        movq mm1, [esi]
        punpcklbw mm1, [dctzero]
        psubw   mm0, [dct128]
        psubw   mm1, [dct128]
        pmaddwd mm0, [dctmatrix+%1]         
        pmaddwd mm1, [dctmatrix+8+%1]
        paddd mm0, mm1    
        add esi, %2              
        
        ; sixth line * first matrix row
        movq mm2, [esi]
        punpckhbw mm2, [dctzero]        
        movq mm3, [esi]
        punpcklbw mm3, [dctzero]
        psubw   mm2, [dct128]
        psubw   mm3, [dct128]
        pmaddwd mm2, [dctmatrix+%1]
        pmaddwd mm3, [dctmatrix+8+%1]
        paddd mm2, mm3    
        add esi, %2              

        ; seventh line * first matrix row
        movq mm4, [esi]
        punpckhbw mm4, [dctzero]        
        movq mm5, [esi]
        punpcklbw mm5, [dctzero]
        psubw   mm4, [dct128]
        psubw   mm5, [dct128]
        pmaddwd mm4, [dctmatrix+%1]
        pmaddwd mm5, [dctmatrix+8+%1]
        paddd mm4, mm5    
        add esi, %2              
        
        ; eightth line * first matrix row
        movq mm6, [esi]
        punpckhbw mm6, [dctzero]        
        movq mm7, [esi]
        punpcklbw mm7, [dctzero]
	psubw   mm6, [dct128]
        psubw   mm7, [dct128]
        pmaddwd mm6, [dctmatrix+%1]
        pmaddwd mm7, [dctmatrix+8+%1]
        paddd mm6, mm7    
                
        movq mm1, mm0                   
        punpckhdq mm0, mm2              
        punpckldq mm1, mm2              
        paddd mm0, mm1                  
        paddd mm0, [dctround]
        psrad mm0, 16

        movq mm3, mm4                   
        punpckhdq mm4, mm6
        punpckldq mm3, mm6
        paddd mm4, mm3                  
        paddd mm4, [dctround]
        psrad mm4, 16
        
        packssdw mm0, mm4               

        movq [edi + 8 + %1], mm0
        pop esi

%endm


; *****************************************************************************
; * DCTY functions			                                      *
; *****************************************************************************
; * C prototype:							      *
; *	void DCTY(short int dest[8][8], char const *source, short int sx);    *
; *                                                                           *
; * Function : Performs the DCT of a 8x8 Y block read directly from the image *
; *            and store the result in dest.				      *
; *									      *
; * Side Effects : MMX registers modified				      *
; *****************************************************************************

; * DCT YUV9 	* ;
; * DCT YUV12 * ;
align 16
DCTY_YUV9:
DCTY_YUV12:
        push ebp
        mov ebp, esp
        push esi
        push edi
	push ebx
	
        mov esi, [ebp + 12]              ; Set pointers (esi = source)
        mov edi, dcttemp         

        movzx ebx, word [ebp + 16]
        
        OneLineDCTYExpand_Planar    0, ebx
        OneLineDCTYExpand_Planar   16, ebx
        OneLineDCTYExpand_Planar   32, ebx
        OneLineDCTYExpand_Planar   48, ebx
        OneLineDCTYExpand_Planar   64, ebx
        OneLineDCTYExpand_Planar   80, ebx
        OneLineDCTYExpand_Planar   96, ebx
        OneLineDCTYExpand_Planar  112, ebx

        mov esi, dcttemp            ; Set pointers
        mov edi, [ebp + 8]                  

        OneLineDCT   0
        OneLineDCT  16
        OneLineDCT  32
        OneLineDCT  48
        OneLineDCT  64
        OneLineDCT  80
        OneLineDCT  96
        OneLineDCT 112

	pop ebx
	pop edi
        pop esi
        pop ebp
        ret

; * DCT YUYV * ;
align 16
DCTY_YUYV:
        push ebp
        mov ebp, esp
        push esi
        push edi
	push ebx
	
        mov esi, [ebp + 12]              ; Set pointers (esi = source)
        mov edi, dcttemp         

        movzx ebx, word [ebp + 16]
        
        OneLineDCTYExpand_YUYV    0, ebx
        OneLineDCTYExpand_YUYV   16, ebx
        OneLineDCTYExpand_YUYV   32, ebx
        OneLineDCTYExpand_YUYV   48, ebx
        OneLineDCTYExpand_YUYV   64, ebx
        OneLineDCTYExpand_YUYV   80, ebx
        OneLineDCTYExpand_YUYV   96, ebx
        OneLineDCTYExpand_YUYV  112, ebx

        mov esi, dcttemp            ; Set pointers
        mov edi, [ebp + 8]                  

        OneLineDCT   0
        OneLineDCT  16
        OneLineDCT  32
        OneLineDCT  48
        OneLineDCT  64
        OneLineDCT  80
        OneLineDCT  96
        OneLineDCT 112

	pop ebx
	pop edi
        pop esi
        pop ebp
        ret


; *****************************************************************************
; * DCTC functions			                                      *
; *****************************************************************************
; * C prototype:							      *
; *	void DCTC(short int dest[8][8], char const *source, short int sx);    *
; *                                                                           *
; * Function : Performs the DCT of a 8x8 Cr or Cb block read directly from    *
; * 	       the image and store the result in dest.			      *
; *									      *
; * Side Effects : MMX registers modified				      *
; *****************************************************************************

; * DCT YUV9 	* ;
align 16
DCTC_YUV9:
        push ebp
        mov ebp, esp
        push esi
        push edi
	push ebx

        mov esi, [ebp + 12]              ; Set pointers (esi = source)
        mov edi, dcttemp         

        movzx ebx, word [ebp + 16]
        
        OneLineDCTCExpand_YUV9    0, ebx
        OneLineDCTCExpand_YUV9   16, ebx
        OneLineDCTCExpand_YUV9   32, ebx
        OneLineDCTCExpand_YUV9   48, ebx
        OneLineDCTCExpand_YUV9   64, ebx
        OneLineDCTCExpand_YUV9   80, ebx
        OneLineDCTCExpand_YUV9   96, ebx
        OneLineDCTCExpand_YUV9  112, ebx

        mov esi, dcttemp            ; Set pointers
        mov edi, [ebp + 8]                  

        OneLineDCT   0
        OneLineDCT  16
        OneLineDCT  32
        OneLineDCT  48
        OneLineDCT  64
        OneLineDCT  80
        OneLineDCT  96
        OneLineDCT 112

	pop ebx
        pop edi
        pop esi
        pop ebp
        ret

; * DCT YUV12 * ;
align 16
DCTC_YUV12:
        push ebp
        mov ebp, esp
        push esi
        push edi
	push ebx

        mov esi, [ebp + 12]              ; Set pointers (esi = source)
        mov edi, dcttemp         

        movzx ebx, word [ebp + 16]
        
        OneLineDCTCExpand_YUV12    0, ebx
        OneLineDCTCExpand_YUV12   16, ebx
        OneLineDCTCExpand_YUV12   32, ebx
        OneLineDCTCExpand_YUV12   48, ebx
        OneLineDCTCExpand_YUV12   64, ebx
        OneLineDCTCExpand_YUV12   80, ebx
        OneLineDCTCExpand_YUV12   96, ebx
        OneLineDCTCExpand_YUV12  112, ebx

        mov esi, dcttemp            ; Set pointers
        mov edi, [ebp + 8]                  

        OneLineDCT   0
        OneLineDCT  16
        OneLineDCT  32
        OneLineDCT  48
        OneLineDCT  64
        OneLineDCT  80
        OneLineDCT  96
        OneLineDCT 112

	pop ebx
        pop edi
        pop esi
        pop ebp
        ret

; * DCT YUYV * ;
align 16
DCTC_YUYV:
        push ebp
        mov ebp, esp
        push esi
        push edi
	push ebx

        mov esi, [ebp + 12]              ; Set pointers (esi = source)
        mov edi, dcttemp         

        movzx ebx, word [ebp + 16]
        
        OneLineDCTCExpand_YUYV    0, ebx
        OneLineDCTCExpand_YUYV   16, ebx
        OneLineDCTCExpand_YUYV   32, ebx
        OneLineDCTCExpand_YUYV   48, ebx
        OneLineDCTCExpand_YUYV   64, ebx
        OneLineDCTCExpand_YUYV   80, ebx
        OneLineDCTCExpand_YUYV   96, ebx
        OneLineDCTCExpand_YUYV  112, ebx

        mov esi, dcttemp            ; Set pointers
        mov edi, [ebp + 8]                  

        OneLineDCT   0
        OneLineDCT  16
        OneLineDCT  32
        OneLineDCT  48
        OneLineDCT  64
        OneLineDCT  80
        OneLineDCT  96
        OneLineDCT 112

	pop ebx
        pop edi
        pop esi
        pop ebp
        ret

; This is only a test for speed
align 16
DCT_TEST:
        push ebp
        mov ebp, esp
        push esi
        push edi
	push ebx

        mov esi, [ebp + 12]              ; Set pointers (esi = source)
        mov edi, dcttemp         

        movzx ebx, word [ebp + 16]

;        OneLineDCTCExpand_YUYV    0, ebx
;        OneLineDCTCExpand_YUYV   16, ebx
;        OneLineDCTCExpand_YUYV   32, ebx
;        OneLineDCTCExpand_YUYV   48, ebx
;        OneLineDCTCExpand_YUYV   64, ebx
;        OneLineDCTCExpand_YUYV   80, ebx
;        OneLineDCTCExpand_YUYV   96, ebx
;        OneLineDCTCExpand_YUYV  112, ebx

        OneLineDCT   0
        OneLineDCT  16
        OneLineDCT  32
        OneLineDCT  48
        OneLineDCT  64
        OneLineDCT  80
        OneLineDCT  96
        OneLineDCT 112

        mov esi, dcttemp            ; Set pointers
        mov edi, [ebp + 8]                  

        OneLineDCT   0
        OneLineDCT  16
        OneLineDCT  32
        OneLineDCT  48
        OneLineDCT  64
        OneLineDCT  80
        OneLineDCT  96
        OneLineDCT 112

	pop ebx
        pop edi
        pop esi
        pop ebp
        ret

END

