;------------------------------------------------------------------------------
;	Fast Assembly MPEG Encoder
;------------------------------------------------------------------------------
; quantize.nas
;	Does the quantization.
;
; Note:	It is likely that this operation could be done during the DCT
;------------------------------------------------------------------------------
;
;	Copyright (C) 1999-2000 Vivien Chappelier
;
; *  This program is free software; you can redistribute it and/or modify
; *  it under the terms of the GNU General Public Licensse as published by
; *  the Free Software Foundation; either version 2 of the License, or
; *  (at your option) any later version.
; *
; *  This program is distributed in the hope that it will be useful,
; *  but WITHOUT ANY WARRANTY; without even the implied warranty of
; *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; *  GNU General Public License for more details.
; *
; *  You should have received a copy of the GNU General Public License
; *  along with this program; if not, write to the Free Software
; *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;
;------------------------------------------------------------------------------

; Macro definitions for portability

%macro cglobal 1 
       global _%1 
%define %1 _%1 
%endmacro

%macro cextern 1 
          extern _%1 
%define %1 _%1 
%endmacro

%ifdef _WIN32_
%define public cglobal
%endif

%ifdef _LINUX_
%define public global
%endif

; /////////////////////////////////////////////////////////////////////////////
; //                         	     DATA                                    //
; /////////////////////////////////////////////////////////////////////////////

section .data
alignb 16

        ; This is the scaling matrix:
	; lowest frequency are coded with more bits than highest
        
        qscale2x  dw 4000h, 2000h, 1AF2h, 1745h, 13B1h, 12F6h, 11A7h, 0F0Fh
                  dw 2000h, 2000h, 1745h, 1555h, 12F6h, 11A7h, 0F0Fh, 0DD6h
                  dw 1AF2h, 1745h, 13B1h, 12F6h, 11A7h, 0F0Fh, 0F0Fh, 0D79h
                  dw 1745h, 1745h, 13B1h, 12F6h, 11A7h, 0F0Fh, 0DD6h, 0CCCh
                  dw 1745h, 13B1h, 12F6h, 11A7h, 1000h, 0EA0h, 0CCCh, 0AAAh
                  dw 13B1h, 12F6h, 11A7h, 1000h, 0EA0h, 0CCCh, 0AAAh, 08B3h
                  dw 13B1h, 12F6h, 11A7h, 0F0Fh, 0D79h, 0B21h, 0924h, 076Bh
                  dw 12F6h, 11A7h, 0EA0h, 0D79h, 0B21h, 0924h, 076Bh, 062Bh

        qround    dd 00010001h
		  dd 00010001h

; /////////////////////////////////////////////////////////////////////////////
; //                         	     CODE                                    //
; /////////////////////////////////////////////////////////////////////////////

section .text
	
        ; The Quantizer (Reduces bits used => loss of quality)
        public QUANTIZE

; *****************************************************************************
; * QuantizeLine macro			                                      *
; *****************************************************************************
; * Function : Read a column from input, performs quantization and store the  *
; * 	       data at the same place.					      *
; *									      *
; * Side Effects : MMX registers modified				      *
; *****************************************************************************
        
%macro QuantizeLine 1

        movq mm0, [esi + %1 + 8]
        movq mm1, [esi + %1]
        pmulhw mm0, [edi + %1 + 8]
        pmulhw mm1, [edi + %1]
        paddw mm0, [qround]
        paddw mm1, [qround]
        psraw mm0, 1
        psraw mm1, 1
        movq [esi + %1 + 8], mm0
        movq [esi + %1], mm1

%endm

; *****************************************************************************
; * QUANTIZE function			                                      *
; *****************************************************************************
; * C prototype:							      *
; *	void QUANTIZE(short int block[8][8], short int qmatrix[8][8]);	      *
; *                                                                           *
; * Function : Performs quantization on a 8x8 block			      *
; *									      *
; * Side Effects : MMX registers modified				      *
; *****************************************************************************

align 16
QUANTIZE:
        push ebp
        mov ebp, esp
        push esi
	push edi

        mov esi, [ebp + 8]
	mov edi, [ebp + 12]
        QuantizeLine   0
        QuantizeLine  16
        QuantizeLine  32
        QuantizeLine  48
        QuantizeLine  64
        QuantizeLine  80
        QuantizeLine  96
        QuantizeLine 112
        
	pop edi
        pop esi
        pop ebp
        ret
END
