;------------------------------------------------------------------------------
;	Fast Assembly MPEG Encoder
;------------------------------------------------------------------------------
; encoding.nas:
;	Performs compression and output of data
;
;------------------------------------------------------------------------------
;
;	Copyright (C) 1999-2000 Vivien Chappelier
;
; *  This program is free software; you can redistribute it and/or modify
; *  it under the terms of the GNU General Public Licensse as published by
; *  the Free Software Foundation; either version 2 of the License, or
; *  (at your option) any later version.
; *
; *  This program is distributed in the hope that it will be useful,
; *  but WITHOUT ANY WARRANTY; without even the implied warranty of
; *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; *  GNU General Public License for more details.
; *
; *  You should have received a copy of the GNU General Public License
; *  along with this program; if not, write to the Free Software
; *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;
;------------------------------------------------------------------------------

; Macro definitions for portability
	
%macro cglobal 1 
       global _%1 
%define %1 _%1 
%endmacro

%macro cextern 1 
          extern _%1 
%define %1 _%1 
%endmacro

%ifdef _WIN32_
%define public cglobal
%endif

%ifdef _LINUX_
%define public global
%endif

; /////////////////////////////////////////////////////////////////////////////
; //                         	     DATA                                    //
; /////////////////////////////////////////////////////////////////////////////
section .data

;------------------------ DC Encoding -----------------------------------------

; The DC encoding tables
alignb 16
%include "video/dcenctab.inc"

;----------------------- RLE-Huffman encoding ---------------------------------


; The Zigzaging table.
align 16
RleHuffZAG  dd   0,  1,  8, 16,  9,  2,  3, 10
            dd  17, 24, 32, 25, 18, 11,  4,  5
            dd  12, 19, 26, 33, 40, 48, 41, 34
            dd  27, 20, 13,  6,  7, 14, 21, 28
            dd  35, 42, 49, 56, 57, 50, 43, 36
            dd  29, 22, 15, 23, 30, 37, 44, 51
            dd  58, 59, 52, 45, 38, 31, 39, 46
            dd  53, 60, 61, 54, 47, 55, 62, 63

; The RLE - Huffman static encoding tables.
HUFFMAXRUN equ 32

align 16
RleHuffMaxLevel dw  41, 19, 6, 5, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2

align 16
%include "video/rlehuff.inc"


; /////////////////////////////////////////////////////////////////////////////
; //                         	     CODE                                    //
; /////////////////////////////////////////////////////////////////////////////
	
section .text

        ; DC Encoding
        public ENCODEYDC
        public ENCODECDC        
        
        ; Zigzag, then compress using RLE and Huffman encoding
        public RLEHUFFZIG 

%include "io/bitwrite.inc"

;------------------------------- DC encoding ----------------------------------

; *****************************************************************************
; * ENCODEYDC function			                                      *
; *****************************************************************************
; * C prototype:							      *
; *	void ENCODEYDC(bitbuffer_t * buffer, short int dcterm, int *predterm) *
; *                                                                           *
; * Function : Encodes the DC component of a Y block according to the DC      *
; *	       encoding tables.						      *
; *									      *
; * Side Effects : None							      *
; *****************************************************************************

align 16
ENCODEYDC:

        push ebp
        mov ebp, esp
        push ebx
        push edi

        movsx  ebx,  word [ebp + 12]	; ebx = dcterm
        mov    edi,  [ebp + 16]		; edi = predterm
	mov    ebp,  [ebp + 8]		; ebp = buffer
        sub    ebx,  [edi]

	mov eax, 7efeh 
	mov ecx, 255 
        cmp ebx, ecx
        jg  encysat
	mov eax, 7e00h
	mov ecx, -255
        cmp ebx, ecx
        jnge encysat

        add  [edi], ebx
        
        add ebx, 255
        shl ebx, 3
        mov eax, [_EncodeYDC_Tab + ebx]
        mov ecx, [_EncodeYDC_Tab + 4 + ebx]
        BitWriteNoMask

        pop edi
        pop ebx
        pop ebp
        ret

align 16
encysat:
        add dword [edi], ecx
        mov ecx, 15
        BitWriteNoMask

        pop edi
        pop ebx
        pop ebp
        ret


; *****************************************************************************
; * ENCODECDC function			                                      *
; *****************************************************************************
; * C prototype:							      *
; *	void ENCODECDC(bitbuffer_t * buffer, short int dcterm, int *predterm) *
; *                                                                           *
; * Function : Encodes the DC component of a Cb or Cr block according to the  *
; *	       DC encoding tables.					      *
; *									      *
; * Side Effects : None							      *
; *****************************************************************************
       
align 16
ENCODECDC:
        push ebp
        mov ebp, esp
        push ebx
        push edi

        movsx  ebx,  word [ebp + 12]	; ebx = dcterm
        mov    edi,  [ebp + 16]		; edi = predterm
	mov    ebp,  [ebp + 8]		; ebp = buffer
        sub    ebx,  [edi]

	mov eax, 0fefeh 
	mov ecx, 255 
        cmp ebx, ecx
        jg  enccsat
	mov eax, 0fe00h
	mov ecx, -255
        cmp ebx, ecx
        jnge enccsat

        add  [edi], ebx
        
        add ebx, 255
        shl ebx, 3
        mov eax, [_EncodeCDC_Tab + ebx]
        mov ecx, [_EncodeCDC_Tab + 4 + ebx]
        BitWriteNoMask

        pop edi
        pop ebx
        pop ebp
        ret

align 16
enccsat:
        add dword [edi], ecx
        mov ecx, 15
        BitWriteNoMask

        pop edi
        pop ebx
        pop ebp
        ret        
;------------------------------ RLE-Huffman encoding --------------------------

; *****************************************************************************
; * RLEHUFFZIG function			                                      *
; *****************************************************************************
; * C prototype:							      *
; * 	void RLEHUFFZIG(bitbuffer_t * buffer, short int block[8][8]);	      *
; *                                                                           *
; * Function : Encodes the AC components of a block according to the	      *
; *	       RLE - Huffman encoding tables.				      *
; *									      *
; * Side Effects : None							      *
; *****************************************************************************

align 16
RLEHUFFZIG:

        push ebp
        mov ebp, esp
        push esi
	push edi
        push ebx
       
        mov esi, [ebp + 12]		; esi = block
        mov ebp, [ebp + 8]		; ebp = buffer
        mov edx, 0
        mov ecx, -1
align 16
RleHuffloop0:
        inc edx
        cmp edx, 64
        jge near RleHuffEnd
        inc ecx
        mov ebx, edx
        mov ebx, [RleHuffZAG + 4*ebx]
        movsx eax, word [esi + 2*ebx]
        or ax, ax
        jz RleHuffloop0
        
        ; ecx = nzeroes
        cmp ecx, HUFFMAXRUN
        jge near RleHuffEscape
        mov ebx, ecx
        movsx ebx, word [RleHuffMaxLevel + 2*ebx]
        cmp ax, bx
        jge near RleHuffEscape
        neg ebx
        cmp ax, bx
        jle RleHuffEscape
align 16
RleHuffHuffman:
        sub eax, ebx
        mov ebx, ecx
        dec eax        
        mov edi, [huff_table + 4*ebx]
        mov ecx, [huff_bits + 4*ebx]
        mov ebx, eax
        mov eax, [edi + 4*ebx]
        movzx ecx, word [ecx + 2*ebx]
        push edx
        BitWriteNoMask
        pop  edx
        mov ecx, -1
        jmp RleHuffloop0

align 16
RleHuffEscape:
        push eax
        push edx
        
        mov eax, 40h
        add eax, ecx
        mov ecx, 12
        BitWriteNoMask          ; ESCAPE + RunLength
        
        pop edx
        pop eax
         
        mov ebx, eax
        add ebx, 255
        movzx eax, word [_RleHuffRLETab + 4*ebx]
        movzx ecx, word [_RleHuffRLETab + 2 + 4*ebx]
        push edx
        BitWriteNoMask
        pop edx

        mov ecx, -1
        jmp RleHuffloop0
align 16
RleHuffEnd:
        mov eax, 2
        mov ecx, 2
        BitWriteNoMask
       
        pop ebx
	pop edi
        pop esi
        pop ebp
        ret

END
